"""
Provides basic token classes for the GiPSy scanner.

Library Version 1.1

Copyright 2013 Paul Griffiths
Email: mail@paulgriffiths.net

Distributed under the terms of the GNU General Public License.
http://www.gnu.org/licenses/
"""


import cgi


class Token:

    """
    Base Token class.

    Class attributes:
    dec_open  : opening token decorator for marked-up output.
    dec_close : closing token decorator for marked-up output
    prefix    : indentation string for outputting in tree view.

    dec_open and dec_close are intended to be directly set by a parser
    or scanner, to implement class-wide decorator strings.

    prefix is not used by the base class, but can be used by subclasses
    (e.g. the container class).

    Instance attributes:
    _token_type_string : a string containing a description of the token type
    _value : the original, unmodified token string

    Public methods:
    __init__()
    read()

    """

    dec_open = ""
    dec_close = ""
    prefix = ""
    tabs = None

    def __init__(self, value):

        """
        Class initializer.

        Arguments:
        value -- the string to be tokenized.
        """

        self._token_type_string = "Unknown token"
        if Token.tabs:
            self._value = value.expandtabs(Token.tabs)
        else:
            self._value = value

    def read(self, decorated=False, html=False, tree=False):

        """
        Returns a string representing the unmodified or modified token.

        Arguments:
        decorated -- set to True to decorate with dec_open and dec_close
        html -- set to True to HTML-escape the characters
        tree -- set to True to output in an indented tree view
        """

        if tree:
            return "%s%s: %s\n" % (Token.prefix,
                                   self._token_type_string,
                                   self._value)
        elif not decorated and not html:
            return self._value
        elif not decorated and html:
            return cgi.escape(self._value, True)
        elif decorated and not html:
            return "%s%s%s" % (self.__class__.dec_open,
                               self._value,
                               self.__class__.dec_close)
        else:
            return "%s%s%s" % (self.__class__.dec_open,
                               cgi.escape(self._value, True),
                               self.__class__.dec_close)

    def get_value(self):

        """
        Returns the value of the _value attribute.
        """

        return self._value


class WhitespaceToken(Token):

    """
    Generic whitespace token class.
    """

    def __init__(self, value):

        """
        Class initializer.

        Arguments:
        value -- the string to be tokenized.
        """

        Token.__init__(self, value)
        self._token_type_string = "Whitespace"

    def read(self, decorated=False, html=False, tree=False):

        # Override superclass function to suppress annoying
        # output of whitespace tokens in the tree view

        if tree:
            return ""
        else:
            return Token.read(self, decorated=decorated, html=html, tree=tree)


class NewlineToken(Token):

    """
    Generic new line token class.
    """

    def __init__(self, value):

        """
        Class initializer.

        Arguments:
        value -- the string to be tokenized.
        """

        Token.__init__(self, value)
        self._token_type_string = "Newline"

    def read(self, decorated=False, html=False, tree=False):

        # Override superclass function to suppress annoying
        # output of whitespace tokens in the tree view

        if tree:
            return ""
        else:
            return Token.read(self, decorated=decorated, html=html, tree=tree)
